<?php

namespace Drupal\component_schema_ui_patterns\Plugin\Deriver;

use Drupal\component_schema\Component\TypedComponentManager;
use Drupal\component_schema\Component\Schema\ComponentAttribute;
use Drupal\component_schema\Component\Schema\ComponentMapping;
use Drupal\component_schema\Component\Schema\ComponentVariableAttributeProviderDataDefinition;
use Drupal\component_schema\Component\Schema\ComponentVariableDataDefinitionInterface;
use Drupal\Core\Config\Schema\Mapping;
use Drupal\Core\Config\Schema\Sequence;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\TypedData\Plugin\DataType\BooleanData;
use Drupal\Core\TypedData\Plugin\DataType\StringData;
use Drupal\Core\TypedData\TypedDataManager;
use Drupal\ui_patterns\Plugin\Deriver\AbstractPatternsDeriver;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class ComponentSchemaDeriver.
 */
class ComponentSchemaDeriver extends AbstractPatternsDeriver {

  /**
   * The typed component manager.
   *
   * @var \Drupal\component_schema\Component\TypedComponentManager
   */
  protected $typedComponentManager;

  /**
   * The component schema wrapper object for the current component object.
   *
   * @var \Drupal\Core\Config\Schema\Element
   */
  protected $schema;

  /**
   * ComponentSchemaDeriver constructor.
   *
   * @param string $base_plugin_id
   *   The base plugin ID.
   * @param \Drupal\Core\TypedData\TypedDataManager $typed_data_manager
   *   Typed data manager service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   Messenger service.
   * @param \Drupal\component_schema\Component\TypedComponentManager $typed_component_manager
   *   The typed component manager.
   */
  public function __construct($base_plugin_id, TypedDataManager $typed_data_manager, MessengerInterface $messenger, TypedComponentManager $typed_component_manager) {
    parent::__construct($base_plugin_id, $typed_data_manager, $messenger);
    $this->typedComponentManager = $typed_component_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, $base_plugin_id) {
    return new static(
      $base_plugin_id,
      $container->get('typed_data_manager'),
      $container->get('messenger'),
      $container->get('component_schema.typed')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getPatterns() {
    $patterns = [];

    foreach ($this->typedComponentManager->getComponentDefinitions() as $component_type => $component) {
      if ($pattern_array = $this->getPatternDefinitionArray($component_type)) {
        if ($pattern = $this->getPatternDefinition($pattern_array)) {
          $patterns[] = $pattern;
        }
      }
    }

    return $patterns;
  }

  /**
   * Returns a pattern definition array based on the given component definition.
   *
   * @param string $component_type
   *   The component type machine name.
   *
   * @return array|FALSE
   *   A pattern definition array or FALSE if the component type doesn't support
   *   an admin UI.
   */
  protected function getPatternDefinitionArray($component_type) {
    $this->schema = $this->typedComponentManager
      ->createFromNameAndData($component_type, []);
    $component_definition = $this->schema->getDataDefinition();
    if ($component_definition->takesUi()) {
      $pattern_definition = [
        // Use the component type machine name as the pattern key.
        'id' => $component_type,
      ];

      // base_path and file_name are required properties of a pattern definition.
      // In our case they're not particularly relevant. The following call uses`
      // the template.
      // @todo If continuing to use template file here, parse the relative path.
      // @todo Consider switching to using the variables_yml file.
      $template_path = $this->schema
        ->getTypedDataManager()
        ->getTwigEnvironment()
        ->load($component_definition->getComponentTemplate())
        ->getSourceContext()
        ->getPath();
      $pattern_definition['base path'] = dirname($template_path);
      $pattern_definition['file name'] = basename($template_path);

      // Copy over equivalent properties.
      // Keys are source (component schema) methods, values are destination
      // (pattern) keys.
      $mapping = [
        'getLabel' => 'label',
        'getDescription' => 'description',
        'getLibraries' => 'libraries',
        'getProvider' => 'provider',
        'getComponentTemplate' => 'use',
      ];
      foreach ($mapping as $source_method => $destination_key) {
        if ($source_value = $component_definition->{$source_method}()) {
          $pattern_definition[$destination_key] = $source_value;
        }
      }

      // Process variables, which will be parsed as pattern fields or settings.
      $elements = $this->schema->getElements();
      foreach (array_keys($elements) as $key) {
        $this->processVariableItem($pattern_definition, $key);
      }

      return $pattern_definition;
    }

    return FALSE;
  }

  /**
   * Parses fields and settings from component variables.
   *
   * In this method we derive a set of fields and/or settings from the
   * component's variables. To do so, we need to convert our nested structure
   * into flat arrays.
   *
   * @param array[] &$pattern_definition
   *   The pattern definition array.
   * @param string $key
   *   The component variable key.
   */
  protected function processVariableItem(array &$pattern_definition, $key) {
    $element = $this->schema
      ->get($key);

    // @todo Support Sequence elements.
    if ($element instanceof Sequence) {
      return;
    }

    $definition = $element->getDataDefinition();

    // Iterate through a mapping.
    if ($element instanceof Mapping) {
      $child_elements = $element->getElements();
      foreach ($child_elements as $child_key => $child_element) {
        $child_definition = $child_element->getDataDefinition();
        // Prepend child variable labels with the parent label.
        $child_definition->setLabel($definition->getLabel() . ' » ' . $child_definition->getLabel());
        $this->processVariableItem($pattern_definition, $key . '.' . $child_key);
      }
      return;
    }

    // Return early if the element doesn't take a UI.
    // @todo: move this test before the handling of sequences and mappings.
    if (!$definition->takesUi()) {
      return;
    }

    // Construct a key using replacements.
    $variable_key = str_replace('.', '__', $key);

    $item = [];

    // Copy over equivalent properties that are used by both fields and
    // settings.
    // Keys are source (component schema) methods, values are destination
    // (field or setting) keys.
    $mapping = [
      'getLabel' => 'label',
      'getDescription' => 'description',
      'getPreview' => 'preview',
    ];
    foreach ($mapping as $source_method => $destination_key) {
      if ($source_value = $definition->{$source_method}()) {
        $item[$destination_key] = $source_value;
      }
    }

    $collection = NULL;

    switch ($definition->getUiType()) {
      case ComponentVariableDataDefinitionInterface::UI_TYPE_FIELD:
        $collection = 'fields';
        $item['type'] = 'string';
        break;
      case ComponentVariableDataDefinitionInterface::UI_TYPE_SETTING:
        $collection = 'settings';
        if ($element instanceof BooleanData) {
          $collection = 'settings';
          $item['type'] = 'boolean';
        }
        // Attribute providers are best modelled as settings.
        elseif ($definition instanceof ComponentVariableAttributeProviderDataDefinition) {
          if ($options = $definition->getOptions()) {
            $item['type'] = 'select';
            $item['options'] = $options;
          }
          else {
            $item['type'] = 'textfield';
          }
        }
        elseif ($element instanceof ComponentAttribute) {
          $item['type'] = 'attributes';
        }
        break;
    }

    if ($collection) {
      $pattern_definition[$collection][$variable_key] = $item;
    }

  }

}
