<?php

namespace Drupal\component_schema\Component;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Config\StorageInterface;
use Drupal\Core\Config\TypedConfigManager;
use Drupal\Core\DependencyInjection\ClassResolverInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Template\TwigEnvironment;

/**
 * Extends TypedConfigManager for use with component template data.
 */
class TypedComponentManager extends TypedConfigManager implements TypedComponentManagerInterface {

  /**
   * The Twig environment.
   *
   * @var \Drupal\Core\Template\TwigEnvironment
   */
  protected $twig;

  /**
   * Creates a new typed configuration manager.
   *
   * @param \Drupal\Core\Config\StorageInterface $configStorage
   *   The storage object to use for reading schema data
   * @param \Drupal\Core\Config\StorageInterface $schemaStorage
   *   The storage object to use for reading schema data
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   The cache backend to use for caching the definitions.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   * @param \Drupal\Core\DependencyInjection\ClassResolverInterface $class_resolver
   *   The class resolver.
   */
  public function __construct(StorageInterface $configStorage, StorageInterface $schemaStorage, CacheBackendInterface $cache, ModuleHandlerInterface $module_handler, ClassResolverInterface $class_resolver) {
    parent::__construct($configStorage, $schemaStorage, $cache, $module_handler, $class_resolver);

    $this->setCacheBackend($cache, 'typed_component_definitions');
    $this->alterInfo('component_schema_info');
  }

  /**
   * {@inheritdoc}
   */
  public function getComponentDefinitions() {
    $definitions = $this->getDefinitions();
    // The definitions include base types. We only want the component types.
    $definitions = array_filter($definitions, function($definition) {
      return !empty($definition['type']) && ($definition['type'] === 'type_mapping');
    });
    return $definitions;
  }

  /**
   * {@inheritdoc}
   */
  public function setTwigEnvironment(TwigEnvironment $twig) {
    $this->twig = $twig;
  }

  /**
   * {@inheritdoc}
   */
  public function getTwigEnvironment() {
    return $this->twig;
  }

  /**
   * {@inheritdoc}
   */
  public function handleTwigDebug() {
    // If in debug mode, we don't want caching.
    if ($this->twig->isDebug()) {
      $this->clearCachedDefinitions();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function clearCachedDefinitions() {
    // Parent calls $this->schemaStorage->reset(), a method that doesn't
    // exist on our storage.
    // Therefore, fork the ::clearCachedDefinitions() code from all ancestors
    // except the parent.
    $this->prototypes = [];
    if ($this->cacheBackend) {
      if ($this->cacheTags) {

        // Use the cache tags to clear the cache.
        Cache::invalidateTags($this->cacheTags);
      }
      else {
        $this->cacheBackend
          ->delete($this->cacheKey);
      }
    }
    $this->definitions = NULL;
  }

}
